#pragma once

#include <iostream>
#include <string_view>
#include <source_location>
#include <cxxabi.h>
#include <libunwind.h>
#include <dlfcn.h>
#include <cstring>
#include <memory>

// Simple RAII wrapper for file descriptor
class FileDescriptor {
    int fd_ = -1;
public:
    explicit FileDescriptor(int fd) : fd_(fd) {}
    ~FileDescriptor() { if (fd_ != -1) close(fd_); }
    operator int() const { return fd_; }
    FileDescriptor(const FileDescriptor&) = delete;
    FileDescriptor& operator=(const FileDescriptor&) = delete;
};

void print_stacktrace() {
    unw_cursor_t cursor;
    unw_context_t context;
    unw_word_t offset, pc;
    char sym[256];
    char *name = sym;
    int status;
    int frame_num = 0;

    // Initialize cursor to current frame
    unw_getcontext(&context);
    unw_init_local(&cursor, &context);

    std::cerr << "Stack trace (most recent call first):\n";

    // Walk up the stack
    while (unw_step(&cursor) > 0) {
        unw_get_reg(&cursor, UNW_REG_IP, &pc);
        if (pc == 0) {
            break;
        }

        std::cerr << "#" << frame_num++ << " ";

        // Get symbol name
        if (unw_get_proc_name(&cursor, sym, sizeof(sym), &offset) == 0) {
            // Demangle the C++ name
            int status = 0;
            size_t length = sizeof(sym);
            name = abi::__cxa_demangle(sym, nullptr, &length, &status);
            
            Dl_info info;
            if (dladdr(reinterpret_cast<void*>(pc), &info) && info.dli_fname) {
                std::cerr << "in " << info.dli_fname << ": ";
            }
            
            if (name) {
                std::cerr << name << " + 0x" << std::hex << offset;
                free(name);
            } else {
                std::cerr << sym << " + 0x" << std::hex << offset;
            }
            
            std::cerr << " [0x" << std::hex << pc << "]\n";
        } else {
            std::cerr << "[unknown] [0x" << std::hex << pc << "]\n";
        }
    }
}

void assert_failed(
    bool condition,
    std::string_view message,
    std::source_location location = std::source_location::current()
) {
    if (!condition) {
        std::cerr << "Assertion failed at " << location.file_name() << ":" << location.line() << ": " 
                  << location.function_name() << "(): " << message << "\n";
        print_stacktrace();
        std::abort();
    }
}

#define ASSERT(condition, message) assert_failed(condition, message, std::source_location::current())
